/*
 *  list.c: part of unbind for Human68k
 *  Copyright (C) 1997  oo<HGG00736@niftyserve.or.jp>
 */
#include <stdio.h>
#include <sys/stat.h>
#include <sys/dos.h>
#include "unbind.h"


static char*str_attribute(dosdir_s* ptr)
{
  int attr = ptr->attr | ptr->fatno;
  static char buff[16];

  buff[0] = ((attr & _DOS_IFDIR)  ? 'd' :
            ((attr & _DOS_IFVOL)  ? 'v' :
            ((attr & _DOS_IFLNK) == _DOS_IFLNK ? 'l' : '-')));
  buff[1] =  (attr & _DOS_IFREG)  ? 'a' : '-';
  buff[2] =  (attr & _DOS_ISYS)   ? 's' : '-';
  buff[3] =  (attr & _DOS_IHIDDEN)? 'h' : '-';
  buff[4] =  'r';
  buff[5] =  (attr & _DOS_IRDONLY) ? '-' : 'w';
  buff[6] =  (attr & _DOS_IEXEC)   ? 'x' : '-';
  buff[7] = '\0';
  return buff;
}


static char*str_datetime(dosdir_s* ptr)
{
  int date, time;
  int yy, mm, dd, h, m, s;
  static char buff[32];

  date = (my_endian ? ptr->date : xshort(ptr->date));
  time = (my_endian ? ptr->time : xshort(ptr->time));

  dd = date & 0x1f;
  mm = (date >> 5) & 0x0f;
  yy = ((date >> 9) & 0x3f) + 1980;

  s  = (time & 0x1f) << 1;
  m  = (time >>  5) & 0x3f;
  h  = (time >> 11) & 0x1f;

  sprintf(buff, "%4d/%02d/%02d %02d:%02d:%02d", yy, mm, dd, h, m, s);
  return buff;
}


void	list(bind_s* root)
{
  int     no;
  bind_s* bp;
  ulong   btabofs = btab_ofs(root->body);

  if (!f_header) {
    fputs(" no    attr      date      time      start     length    filename\n"
          "---  -------  ---------- --------  --------  ----------  -------------\n"
          , stdout);
  }

  for (no = 0, bp = root; bp; bp = bp->next, ++no) {
    printf("%3d  %s  %s  %08lx  %10ld  %s\n",
           no,
           str_attribute(bp->btab),
           str_datetime(bp->btab),
           bp->btab->size,
           (bp->next ? (bp->next)->btab->size : btabofs) - bp->btab->size,
           dosname2filename(bp->btab, NULL));
  }
}


