/*
 *  unbind.c: part of unbind for Human68k
 *  Copyright (C) 1997  oo<HGG00736@niftyserve.or.jp>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/dos.h>
#include "getopt.h"
#include "error.h"
#include "unbind.h"

/*
 *	global variables
 */
int    my_endian;
char   tempfile[PATH_MAX];
char*  bindfile;
bind_s*bindroot;


/*
 *	option variables for GNU getopt()
 */
int f_append  = 0;
int f_check   = 0;
int f_create  = 0;
int f_delete  = 0;
int f_list    = 0;
#if	0
int f_update  = 0;
#endif
int f_extract = 0;

int f_all     = 0;	/* internal flag */
int f_debug   = 0;
int f_header  = 0;
int f_help    = 0;
int f_secret  = 0;
int f_verbose = 0;
int f_version = 0;

static struct option long_opt[] = {
	{"append",	no_argument,	NULL,	'a'},
	{"check",	no_argument,	NULL,	'C'},
	{"create",	no_argument,	NULL,	'c'},
	{"debug",	no_argument,	NULL,	'D'},
	{"delete",	no_argument,	NULL,	'd'},
	{"header",	no_argument,	NULL,	'H'},
	{"help",	no_argument,	NULL,	'h'},
	{"list",	no_argument,	NULL,	'l'},
	{"secret",	no_argument,	NULL,	's'},
	{"table",	no_argument,	NULL,	't'},
#if	0
	{"update",	no_argument,	NULL,	'u'},
#endif
	{"version",	no_argument,	NULL,	'V'},
	{"verbose",	no_argument,	NULL,	'v'},
	{"extract",	no_argument,	NULL,	'x'},
	{0, 0, 0, 0}
};

#define	GETOPT_STRING	"aCcDdHhlstVvx"
#define	HELP_STRING	"-" GETOPT_STRING


static void print_help_and_die(void)
{
  fprintf(stderr,
"usage: %s [" HELP_STRING "] bind_file [file ...]\n"
#ifdef	JMSG
" opts: -a, --append   oCht@Cւ̒ǉ\n"
"       -C, --check    oCht@Č\n"
"       -c, --create   oCht@C̐VK쐬\n"
"       -D, --debug    fobObZ[W̕\\n"
"       -d, --delete   oCht@Ct@C̍폜\n"
"       -H, --header   wb_\̗}\n"
"       -h, --help     ̃bZ[W̕\\n"
"       -l, --list     oCht@C̃Xg\\n"
"       -s, --secret   oCht@CɉB\n"
"       -t, --table    oCht@C̃Xg\\n"
"       -V, --version  o[W̕\\n"
"       -v, --verbose  bZ[W璷ɂ\n"
"       -x, --extract  oCht@Ct@C̒o\n"
#else
" opts: -a, --append   append file(s) to bind_file\n"
"       -C, --check    check bind_file\n"
"       -c, --create   create new bind_file\n"
"       -D, --debug    print debug message\n"
"       -d, --delete   delete file(s) from bind_file\n"
"       -H, --header   disable print header\n"
"       -h, --help     print this message\n"
"       -l, --list     print list of bind .X file\n"
"       -s, --secret   add secret attribute to bindfile\n"
"       -t, --table    print list of bind .X file\n"
"       -V, --version  print version\n"
"       -v, --verbose  verbose mode\n"
"       -x, --extract  extract file(s) from bind_file\n"
#endif
         , progname);
  exit(-1);
}


static void print_version(void)
{
  fprintf(stderr,
          "%s %s  Copyright (C) 1997 oo\n"
          , progname, UNBIND_VERSION);
}


static bind_s* delete(bind_s* root, int ac, char** av)
{
  bind_s* bp;
  char*   name;

  while (ac > optind) {
    name = av[optind++];
    if (root->next == NULL) {
#ifdef	JMSG
      fatal("Ō̃W[ %s ͍폜ł܂", name);
#else
      fatal("cannot delete last module %s", name);
#endif
    }
    if ((bp = delete_module(root, name)) == NULL) {
#ifdef	JMSG
      fatal("oCht@C %s ɃW[ %s ͊܂܂Ă܂", bindfile, name);
#else
      fatal("module %s is not exist in bind_file %s", name, bindfile);
#endif
    }
    else {
      root = bp;
    }
  }
  return root;
}


static bind_s* append(bind_s* root, int ac, char** av)
{
  bind_s* bp;
  char*   name;

  while (ac > optind) {
    name = av[optind++];
    if ((bp = append_module(root, name)) == NULL) {
#ifdef	JMSG
      fatal("oCht@C %s ɃW[ %s ǉł܂ł", bindfile, name);
#else
      fatal("module %s cannot append to bind file %s", name, bindfile);
#endif
    }
    else {
      if (isBfile(bp->body)) {
#ifdef	JMSG
        fatal("oCht@C %s ͒ǉł܂", name);
#else
        fatal("duplicate binding %s", name);
#endif
      }
      if (root == NULL) {
        root = bp;
      }
    }
  }
  return root;
}


#ifndef	HUMAN68K
static void _toslash(char* path)
{
  int   c;
  char* dst;

  for (dst = buf; c = *path++; ) {
    if (c == '\\') c = '/';
    *dst++ = c;
  }
}


static void _splitpath(const char* path, char* drv, char* dir, char* file, char* ext)
{
  char* ptr;
  char* src;
  char* dst;
  char buf[PATH_MAX];

  if ((src = path) == NULL || *path == '\0') {
    drv[0]  = '\0';
    dir[0]  = '\0';
    file[0] = '\0';
    ext[0]  = '\0';
    return;
  }

  dst = drv;
#if  defined(MSDOS) || defined(HUMAN68K)
  if (isalpha(src[0]) && src[1] == ':') {
    *dst++ = toupper(src[0]);
    *dst++ = ':';
    src += 2;
  }
#endif
  *dst = '\0';

#if  defined(MSDOS) || defined(HUMAN68K)
  src = _toslash(strcpy(buf, src));
#endif
  dst = dir;
  if (ptr = strrchr(src, '/')) {
    while ((ptr - src) >= 0) {
      *dst++ = *src++;
    }
  }
  *dst = '\0';

  dst = file;
  if (strcmp(src, ".") == 0 || strcmp(src, "..")) {
    strcpy(file, src);
    ext[0] = '\0';
  }
  else {
    if (ptr = strrchr(src, '.')) {
      while ((ptr - src) > 0) {
        *dst++ = *src++;
      }
    }
    *dst = '\0';

    if (ptr) {
      strcpy(ext, ptr);
    }
    else {
      ext[0] = '\0';
    }
  }
}


static void _makepath(char* path, const char* drv, const char* dir,
                      const char* file, const char* ext)
{
  int  len;
  char*src;
  char buf[PATH_MAX];

  path[0] = '\0';
  if (drv && *drv) {
    strcat(path, drv);
  }

#if  defined(MSDOS) || defined(HUMAN68K)
  dir = (const char*)strcpy(buf, dir);
  _toslash(dir);
#endif
  if (dir && *dir) {
    strcat(path, dir);
    len = strlen(path);
    if (path[len - 1] != '/') {
      path[len++] = '/';
      path[len]   = '\0';
    }
  }

  if (file && *file) {
    strcat(path, file);
  }

  if (ext && *ext) {
    len = strlen(path);
    if (path[len - 1] != '.') {
      path[len++] = '.';
    }
    strcpy(path + len, ext + (ext[0] == '.' ? 1 : 0));
  }
}
#endif


static char* tempname(char* file)
{
  char drv[4], dir[PATH_MAX], nam[32], ext[4];

  _splitpath(bindfile, drv, dir, nam, ext);
#ifdef	HUMAN68K
  sprintf(nam, "unbi%04x", (getpid()>>8)&0xffff);
#else
  sprintf(nam, "unbi%04x", getpid());
#endif
  ext[0] = '\0';
  _makepath(file, drv, dir, nam, ext);
  return file;
}


static void bind(int ac, char** av)
{
  /* e|t@C쐬 */
  if (tempname(tempfile) == NULL) {
#ifdef	JMSG
    fatal("e|t@C蓖Ă܂");
#else
    fatal("cannot make temporary filename");
#endif
  }

  if (f_delete) {
    bindroot = delete(bindroot, ac, av);
  }
  else if (f_create || f_append) {
    bindroot = append(bindroot, ac, av);
  }
  else {
#ifdef	JMSG
    fatal("̑̓T|[gĂ܂");
#else
    fatal("this operation is not support");
#endif
  }
  write_Bfile(bindroot, tempfile);

  /* ̃t@C폜ăt@CύX */
  _dos_chmod(bindfile, _DOS_IFREG);
  if (remove(bindfile) == -1) {
    if (!f_create) {
#ifdef	JMSG
      fatal("oCht@C %s폜ł܂", bindfile);
#else
      fatal("cannot delete bindfile %s", bindfile);
#endif
    }
  }
  if (rename(tempfile, bindfile) == -1) {
#ifdef	JMSG
    fatal("t@C %s  %s ɕύXł܂", tempfile, bindfile);
#else
    fatal("cannot rename %s to %s", tempfile, bindfile);
#endif
  }
#if	0
  if (f_update) {
    _dos_chmod(bindfile, bindroot->btab->attr);
  }
#endif
}


int main(int ac, char** av)
{
  int c, count;
  int ops, opts;

  my_endian = endian();
  set_progname(av[0]);

  while ((c = getopt_long(ac, av, GETOPT_STRING, long_opt, &count)) != EOF) {
    switch (c) {
    case 'a':  f_append = 1;   break;
    case 'C':  f_check  = 1;   break;
    case 'c':  f_create = 1;   break;
    case 'D':  f_debug  = f_verbose = 1;   break;
    case 'd':  f_delete = 1;   break;
    case 'H':  f_header = 1;   break;
    case 'h':  f_help   = 1;   break;
    case 'l':  f_list   = 1;   break;
    case 's':  f_secret = 1;   break;
    case 't':  f_list   = 1;   break;
#if	0
    case 'u':  f_update = 1;   break;
#endif
    case 'V':  f_version = 1;  break;
    case 'v':  f_verbose = 1;  break;
    case 'x':  f_extract = 1;  break;
    default:
      print_help_and_die();
      break;
    }
  }

  /*
   *  IvṼ`FbN
   */
  ops  = f_append + f_check + f_create + f_delete + f_extract;
  opts = ops + f_list + f_help + f_version;

  switch (ops) {
  case 0:
    if (opts == 0) {
      f_list = 1;
    }
    break;
  case 1:
    if (ac == (optind + 1)) {
      if (f_extract) {
        f_all = 1;
      }
      else if (!f_check) {
#ifdef	JMSG
        error("삷t@Cw肵Ă");
#else
        error("specify filename to operate");
#endif
        f_help = 1;
      }
    }
    break;
  default:
#ifdef	JMSG
    error("IvV -a, -C, -c, -d, -x ͈w肵ĉ");
#else
    error("select one option from -a, -C, -c, -d, -x");
#endif
    f_help = 1;
  }

  /*
   *  p[^̐̃`FbN
   */
  if (opts == 0 && ac == optind) {
    f_help = 1;
  }

  if (f_version)    print_version();
  if (f_help)       print_help_and_die();
  if (ac == optind) return 0;

  /* oCht@Cǂݍ */
  bindfile = av[optind++];
  if (!f_create) {
    if ((bindroot = read_Bfile(bindfile)) == NULL) {
      if (f_check) return (errno ? errno : -1);
#ifdef	JMSG
      fatal("oCht@C %s ǂݍ߂܂", bindfile);
#else
      fatal("read error at bindfile %s", bindfile);
#endif
    }
    if (f_check) return 0;
  }

  if (f_extract) {
    if (f_list) list(bindroot);
    extract(bindroot, ac, av);
  }
  else if (ops) {
    bind(ac, av);
    if (f_list) list(bindroot);
  }
  else if (f_list) {
    list(bindroot);
  }

  return 0;
}


